
from wikifile.wikiFile import WikiFile
import json
import sys
import logging

from wikifile.cmdline import CmdLineAble


class WikiExtract(CmdLineAble):
    """
    Provides methods to extract data from wiki markup files.
    """
    def __init__(self, debug=False, is_module_call=False):
        '''
        constructor
        
        Args:
            debug(bool): if True show debugging information
        '''
        self.debug=debug
        self.is_module_call=is_module_call
        super(WikiExtract, self).__init__()

    def getParser(self):
        # Setup argument parser
        super().getParser() # setup default parser arguments
        if self.parser is None:
            raise AttributeError("parser of this object should be defined at this point.")
        self.parser.add_argument("-m", "--mode", dest="mode",
                                 help="Select a mode.\n\tupdate_templates: updates the wikifiles at the provided location with the provided data\n\tcreate: creates a wikifile with the given data.",
                                 required=True)
        # Add parser arguments
        self.parser.add_argument("-t", "--template", dest="template",
                                 help="Select a template in which the data is being rendered", required=True)
        self.parser.add_argument("-id", "--file_name_id", dest="file_name_id",
                                 help="Name of the key in which the file name is stored.")
        self.parser.add_argument("--listFile", dest="file_list",
                                 help="List of pages from which the data should be extracted", required=False)

    @staticmethod
    def extract_templates(template_name: str, stdIn, page_titles, file_list, backup_path, add_file_name):
        """
        Extracts template data of the template template_name from the given files/location and returns the data as json.
        Args:
            template_name: name of the template that should be extracted
            stdIn:
            page_titles:
            file_list:
            backup_path:
            add_file_name: If defined this value will be used as key to store the filename. Should be used wisely to not interfere with regular template arguments.

        Returns:

        """
        # refactor to use command line
        res = []
        for file in page_titles:
            wikiFile = WikiFile(file, backup_path)
            template = wikiFile.extract_template(template_name)
            if template is not None:
                if add_file_name is not None:
                    template[add_file_name] = file
                res.append(template)
        return json.dumps({"data": res}, default=str, indent=3)
    
    def maininstance(self, argv=None):
        '''
        command line entry point
        
        Args:
            argv(list): command line arguments
        '''
        if argv is None:
            argv = sys.argv[1:]

        self.getParser()
        try:
            # Process arguments
            args = self.parser.parse_args(argv)
            super().initLogging(args)


            res_templates = WikiExtract.extract_templates(args.template,
                                                          stdIn=args.stdin,
                                                          file_list=args.file_list,
                                                          page_titles=args.pages,
                                                          backup_path=args.backupPath,
                                                          add_file_name=args.file_name_id)
            print(res_templates)


        except KeyboardInterrupt:
            ### handle keyboard interrupt ###
            return 1
        except Exception as e:
            print(e)

def main_module_call():
    '''
    main entry point for console script
    '''
    wikiextract = WikiExtract(is_module_call=True)
    wikiextract.maininstance()
    sys.exit

if __name__ == '__main__':
    main_module_call()



from argparse import ArgumentParser
from argparse import RawDescriptionHelpFormatter
import os
import sys
import logging

from lodstorage.jsonable import JSONAble, Types


class CmdLineAble(object):
    """
    Bundles methods that are required by the command line tools that this file provides
    """

    def __init__(self):
        pass
        
    def getParser(self):
        # Setup argument parser
        self.parser = ArgumentParser(formatter_class=RawDescriptionHelpFormatter)
        self.parser.add_argument('-p', '--pages', dest="pages",  nargs='+',
                            help="Names of the pages the action should be applied to")
        self.parser.add_argument('--wikiTextPath', dest="backupPath", help="Path to store/update the wiki entries",
                            required=True)
        self.parser.add_argument('-stdin', dest="stdin", action='store_true',
                            help='Use the input from STD IN using pipes')
        self.parser.add_argument('-ex', '--exclude', dest="exclude_keys",
                            help="List of keys that should be excluded")
        self.parser.add_argument('--debug', dest="debug", action='store_true', default=False, help="Enable debug mode")

        
    def initLogging(self,args):
        if args.debug:
            logging.basicConfig(level=logging.DEBUG, stream=sys.stdout)
        else:
            logging.basicConfig(stream=sys.stdout, level=logging.INFO)
                        

    def getPageTitlesForArgs(self,args):
        '''
        see also wikirestore in wikipush of py-3rdparty-mediawiki
        
        Args:
            args(): parsed arguments
            
        Returns:
            List of pageTitles as specified
        '''
        page_titles=args.pages
        stdIn=args.stdin
        file_list=args.file_list
        backup_path=args.backupPath
        file_parameters = [args.stdin, args.pages, args.file_list]
        if len(file_parameters) - (file_parameters.count(None) + file_parameters.count(False)) > 1:
                logging.error(
                    "Multiple file selection options were used. Please use only one or none to select all files in the backup folder.")
                raise Exception("Invalid parameters")
        if stdIn:
            backup_path = os.path.dirname(page_titles[0].strip())
            pageTitlesfix = []
            for i in page_titles:
                pageTitlesfix.append(os.path.basename(i.strip().replace('.wiki', '')))
            page_titles = pageTitlesfix
        elif file_list is not None:
            f = open(file_list, 'r')
            allx = f.readlines()
            page_titles = []
            for i in allx:
                page_titles.append(os.path.basename(i.strip()).replace('.wiki', ''))
        else:
            if backup_path is None:
                logging.warning("No backup path is defined. Please provide a path to the location were the wiki-files are stored.")
            if page_titles is None:
                page_titles = []
                for path, subdirs, files in os.walk(backup_path):
                    for name in files:
                        filename = os.path.join(path, name)[len(backup_path) + 1:]
                        if filename.endswith(".wiki"):
                            page_titles.append(filename[:-len(".wiki")])
        total = len(page_titles)
        logging.debug(f"extracting templates from {total} wikifiles.")
        return page_titles

    def fromJson(self, listname: str, jsonStr: str, sample: list = None):
        """
        Convert the given string to JSON and convert the values to the corresponding datatype"
        Args:
            listname: name of the list the data is stored in
            jsonStr: data as json string that should be converted
            sample: sample data of the actual data. Used to convert the values to the correct type
        Returns:
            
        """""
        types = Types(listname)
        types.getTypes("data", sample, 1)
        json_parsed = JSONAble()
        json_parsed.fromJson(jsonStr, types)
        return json_parsed.__dict__[listname]



from wikifile.wikiFile import WikiFile
from wikibot.wikipush import WikiPush
from wikifile.cmdline import CmdLineAble
from wikifile.wikiRender import WikiRender
from lodstorage.lod import LOD

class WikiFileManager(CmdLineAble):
    '''
    access to Wiki markup files for a given wiki
    '''

    def __init__(self, sourceWikiId:str, login=True,debug=False):
        '''
        constructor
        
        Args:
            fromWikiId(str): the wikiId of the wiki 
            login(bool): do we need to login to the wiki
            debug(bool): True if debugging should be switched on 
        '''
        super(WikiFileManager, self).__init__()
        self.sourceWikiId=sourceWikiId
        self.wikiPush = WikiPush(fromWikiId=sourceWikiId, login=login)
        self.debug = debug
        self.wikiRender = WikiRender()

    def importLODtoWiki(self, data: list, wikiSon: str, titleKey: str = "pageTitle"):
        """
        Uses the given data and updates the corresponding pages in the wiki.
        It is assumed that the page title is part of the dict binding the values of the dict to the corresponding page.

        Args:
            data(list): List of Dicts that should be used to update the corresponding pages in the wiki
            wikiSon(str): Name of the wikiSon object that should be updated/created
            titleKey(str): Name of the key that holds the pageTitle. Default is "pageTitle"

        Returns:

        """
        pageDict = self.pagesListToDict(data, titleKey)
        wiki_files = self.getUpdatedPages(pageDict, wikiSon)
        self.pushWikiFilesToWiki(wiki_files)


    def exportWikiSonToLOD(self, pageTitels: list, wikiSonName: str, pageTitleKey: str = "pageTitle",
                           properties: list = [], limitProperties: bool = False) -> list:
        """
        Exports the given given WikiSon entities corresponding to the given WikiSonName of the pages corresponding to
        the given pageTitles and returns the values as list of dicts.

        Args:
            pageTitles(list): List of all pageTitles from which the given WikiSon entity should be extracted
            wikiSonName(str): Name of the WikiSon object that should be extracted
            pageTitleKey(str): Name of the key that should be used to identify the pageTitle. This name should be distinct form other properties of the object
            properties(list): List of property names that should occur in the returned LoD. Order of the list is used as order of result. Default is null.
            limitProperties(bool): If true the resulting dicts only contain keys that are present in the given properties list any other key is removed. Otherwise all properties that are either given or defined are present in the result. Defualt is False.

        Returns:
            List of dicts containing the WikiSon entities of the given pages
        """
        lod = []
        for pageTitle in pageTitels:
            wikiFile = self.getWikiFile(pageTitle)
            wikiSon = wikiFile.extract_template(wikiSonName)
            if wikiSon is not None:
                wikiSon[pageTitleKey] = pageTitle
                lod.append(wikiSon)

        # Build up the set of keys
        keys = set()
        if not limitProperties:
            for record in lod:
                used_keys = {key for key in record.keys()}
                keys.update(used_keys)

        # Add key to the properties
        for key in keys:
            if key not in properties:
                properties.append(key)

        # clean up the entries by adding missing keys
        for record in lod:
            for key in properties:
                if key not in record:
                    record[key] = None
            if limitProperties:
                recordKeys=list(record.keys())
                for key in recordKeys:
                    if key not in properties:
                        del record[key]

        # order dicts according to the order of the property list
        propertyMap = {key: pos for pos, key in enumerate(properties)}
        for pos, record in enumerate(lod):
            lod[pos] = dict(sorted(record.items(), key=lambda x: propertyMap[x[0]]))
        return lod

    def convertWikiFilesToLOD(self, wikiFiles: list, wikiSonName: str):
        '''
        converts the given pagesTitles to csv by extracting the given templateName from the wikiPage corresponding to
        the given pageTitle.

        Args:
            wikiFiles(list): PageTitles to fetch and convert to CSV
            wikiSonName(str): Name of the WikiSon object that should be extracted

        Returns:
            header(list): Header for CSV File
            pageDicts(List of Dics): List of Dicts of pages
        '''
        lod = []
        for wikifile in wikiFiles:
            if isinstance(wikifile, WikiFile):
                values = wikifile.extract_template(wikiSonName)
                if values is None:
                    values = {}
                pageTitle = wikifile.getPageTitle()
                if pageTitle is not None and len(values)>0:
                    values['pageTitle']= pageTitle
                    lod.append(values)
        return lod

    def pagesListToDict(self, data: list, titleKey: str = "pageTitle", removeKey:bool=True) -> dict:
        """
        Converts the given list of dicts to a dict in which each value is identified by the corresponding pageTitle.
        It is assumed that each dict has the pageTitle key set and that this key is unique within the list.
        Example:
            Input: [{"pageTitle":"Test Page", "label":Test}]
            Output: {"Test Page":{"label":Test}}

        Args:
            data(list): List of dicts of WikiSon values
            titleKey(str): Name of the key that holds the pageTitle. Default is "pageTitle"
            removeKey(bool): if True remove the key entry of each dict

        Returns:
            dict of dicts where each dict is identified by the pageTitle
        """
        pagesDict,duplicates = LOD.getLookup(data,titleKey,withDuplicates=False)
        if len(duplicates)>0:
            raise Exception(f"there are duplicate pageTitles {duplicates}")
        if removeKey:
            for record in pagesDict.values():
                del record[titleKey] 
        return pagesDict

    def pushWikiFilesToWiki(self, wiki_files: list):
        """
        Pushes the content of the given wikiFiles to the corresponding wiki pages in the wiki
        
        Args:
            wiki_files: list of WikiFiles that should be pushed to the wiki

        Returns:
            Nothing
        """
        for wiki_file in wiki_files:
            if isinstance(wiki_file, WikiFile):
                page_content = str(wiki_file)
                update_msg = f"modified through csv import by {self.wikiPush.fromWiki.wikiUser.user}"
                page = wiki_file.getPage()
                if page is None:
                    page = self.wikiPush.fromWiki.getPage(wiki_file.getPageTitle())
                page.edit(page_content, update_msg)

    def getUpdatedPages(self, records: dict, wikiSon: str) -> list:
        """
        Updates the wikiPages with the given values by applying the values to the WikiSon object matching the given wikiSon name.
        The update is applied by creating a WikiFile object from the page content in the wiki.
        The changes are then applied in the WikiFile object.
        Note: At this point the changes are not applied to the wiki. Do do so use pushWikiFilesToWiki()

        Args:
            records(dict): dict of dicts in which the new values for a page are identified by the pageTitle
            wikiSon(str): Name of the wikiSon object that should be updated/created

        Returns:
            List of WikiFile objects with the updated content
        """
        res = []
        for pageTitle, values in records.items():
            if pageTitle is not None:
                wiki_file = self.getUpdatedPage(pageTitle, values, wikiSon)
                res.append(wiki_file)
        return res

    def getUpdatedPage(self, pageTitle: str, values: dict, wikiSon: str) -> WikiFile:
        """
        Updates the page corresponding to the given pageTitle with the given values

        Args:
            pageTitle(str): Title of the page that shoud be updated
            values(dict): values that should be used for the update
            wikiSon(str): Name of the WikiSon that should be updated

        Returns:
            WikiFile corresponding to the given pageTitle with the applied updates
        """
        wiki_file = self.getWikiFile(pageTitle)
        wiki_file.add_template(wikiSon, values, overwrite=True, prettify=True)
        return wiki_file

    def getWikiFile(self, pageTitle: str) -> WikiFile:
        """
        Queries the given page and converts it to a WikiFile.

        Args:
            pageTitle: Title of the page that should be retrieved

        Returns:
            WikiFile corresponding the the given pageTitle
        """
        pageItem = self.wikiPush.fromWiki.getPage(pageTitle)
        wiki_file = WikiFile(pageTitle, "tmp", self.wikiRender, pageItem.text())
        wiki_file.setPage(pageItem)
        return wiki_file


'''
Created on 2021-03-07

@author: wf
'''
from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from wikifile.wikiRender import WikiRender
    from wikifile.metamodel import Topic, Property, UML, Context


class SMWPart(object):
    '''
    a technical Semantic MediaWiki Part
    '''

    def __init__(self, part, wikiRender=None):
        '''
        Constructor
        '''
        self.part = part
        self.wikiRender = wikiRender
        self.template = "%s_page.jinja" % part.lower().replace(" ", "_")

    def render_page(self, topic: Topic):
        """
        Renders the help page for the given entity using the provided properties
        Args:
            topic: topic for which the page should be rendered
            properties: list of all properties
        Returns:

        """
        template_template = self.wikiRender.template_env.get_template(self.template)
        page = template_template.render(topic=topic)
        return page

    @staticmethod
    def getAll(wikiRender: WikiRender):
        smwPartList = [
            ListOf(wikiRender),
            SMWPart("Help"),
            SMWPart("Category"),
            SMWPart("Concept"),
            Form(wikiRender),
            Template(wikiRender),
            #TODO: implement
            #SMWPart("Properties"),
            #SMWPart("PythonCode")
        ]
        smwParts = {}
        for smwPart in smwPartList:
            smwPart.wikiRender = wikiRender
            smwParts[smwPart.part] = smwPart
        return smwParts

    def get_page_name(self, topic: Topic):
        return f"{self.part}:{topic.name}"

    @staticmethod
    def getAllAsPageLink(topic: Topic):
        """Returns all technical pages of the given topic as link list (also known as the see also section)"""
        return SMW.render_as_list([f"[[:{smwPart.get_page_name(topic)}]]" for smwPart in SMWPart.getAll(None).values()])


class SMW:
    """Provides functions covering basic SMW features"""

    @staticmethod
    def parser_function(function_name: str, presence_is_true=False, **kwargs):
        """
        Renders the given parameters and function name to the corresponding SMW parser function.
        Parameter names containing a whitespace must be written with an underscore instead.
        Args:
            function_name: name of the function
            presence_is_true: If true only the name of bool parameters is displayed. Otherwise the bool value is printed out.
            **kwargs: parameters of the parser function

        Returns:

        """
        return "{{#" + function_name + ":" + SMW.render_parameters(presence_is_true=presence_is_true, **kwargs)[1:] + "}}"

    @staticmethod
    def render_entity(template: str, oneliner=True, **kwargs):
        """
        Renders the given parameters as template of the given name.
        Args:
            template: name of the template
            oneliner= If True entity is returned in oneliner. Otherwise the entity is rendered in a prettier format.
            **kwargs: parameters of the template
        Returns:

        Example:
            Args:
                template="Event"
                oneliner= False
                kwargs= 'Title'='SMWCon', 'Year'='2020'
            Returns:
                    {{Event
                    |Title= SMWCon
                    |Year= 2020
                    }}
        """
        separator = "" if oneliner else "\n"
        return "{{" + template + separator + SMW.render_parameters(oneliner=oneliner, **kwargs) + "}}"

    @staticmethod
    def render_sample_entity_with_properties(topic: Topic, properties: list, oneliner=True):
        """

        Args:
            topic: Topic for which the sample entity template should be generated
            properties: properties of the topic
            oneliner: If true the result will be in one line. Otherwise, result string is returned in a prettier format.
        Returns:

        Example:
            Args:
                template="Event"
                properties= [<Title property>, <Year property>]
                oneliner= False
            Returns:
                    {{Event
                    |Title= Some Title
                    |Year= Some Year
                    }}
        """
        property_dict = {}
        for property in properties:
            property_dict = {**property_dict, property.name: f"Some {property.label}"}
        return SMW.render_entity(topic.name, oneliner=oneliner, **property_dict)

    @staticmethod
    def render_parameters(oneliner=True, presence_is_true=False, **kwargs):
        """

        Args:
            oneliner: If true parameters are rendered in one line.
            presence_is_true: If true only the name of bool parameters is displayed. Otherwise the bool value is printed out.
            **kwargs: All paramerters with there values. If a parameter has a whitespace escape it with an underscore.
        Returns:
            Returns the given parameters as rendered mediawiki template parameters
        """
        separator = "" if oneliner else "\n"
        res = ""
        for parameter, value in kwargs.items():
            if isinstance(value, bool):
                label = parameter.replace("_", " ")
                if presence_is_true:
                    if value:
                        res += f"|{label}{separator}"
                else:
                    # ToDo: Update bool values if decided how to query wiki config
                    bool_value = "true" if value else "false"
                    res += f"|{label}={bool_value}{separator}"
            elif value is not None:
                label = parameter.replace("_", " ")
                res += f"|{label}={value}{separator}"
        return res

    @staticmethod
    def set_entity_parameter(topic, properties, oneliner=True, withDescription=False):
        """

        Args:
            topic: topic for which the properties should be set
            properties: properties which should be stored
            oneliner: If true the result will be in one line. Otherwise, result string is returned in a prettier format.

        Returns:
        """
        property_dict = {"isA": topic.name}
        for property in properties:
            property_dict = {**property_dict, property.get_pageTitle(withNamespace=False): "{{{" + property.name + "|}}}"}
        return SMW.parser_function("set", oneliner=oneliner, **property_dict)

    @staticmethod
    def render_as_list(data: list, is_ordered=False, prefix=""):
        """
        Renders the given data as mediawiki list. If a value in the data is also a list a sublist entry is generated.
        Args:
            data: data that should be rendered as list. Can also contain lists.
            is_ordered: If true an ordered list is returned based on the order in the given data. Otherwise, unordered list is returned.
            prefix: string that is placed before each list item
        Returns:

        """
        symbol = prefix
        symbol += "#" if is_ordered else "*"
        res = ""
        for d in data:
            if isinstance(d, list):
                res += SMW.render_as_list(d, is_ordered, symbol)
            else:
                res += f"{symbol}{d}\n"
        return res


class Template(SMWPart):
    """
    Provides methods to generate a Template page for a topic
    """

    def __init__(self, wikiRender=None):
        if wikiRender is not None:
            wikiRender.template_env.globals['Template'] = self
        super().__init__("Template", wikiRender)

    @staticmethod
    def get_page_name(topic: Topic):
        return f"Template:{topic.name}"

    @staticmethod
    def template_arg(arg):
        return "{{{" + arg + "|}}}"

    @staticmethod
    def table_of_arguments(topic: Topic, properties: list, clickable_links=True):
        """
        ToDo
        Args:
            topic:
            properties:

        Returns:

        """
        formlink = Form.formlink(form=topic.name, link_text="✎", target="{{FULLPAGENAME}}", tooltip="Start editing this " + topic.name)
        table_str = "{| class='wikitable'\n! colspan=2 | " + formlink + topic.get_page_link()+ "\n|-\n"
        sortBySortPos = lambda property: 99999 if "sortPos" not in property.__dict__ or property.__dict__.get(
            "sortPos") is None else int(property.sortPos)
        properties_sorted = sorted(properties, key=sortBySortPos)
        for property in properties_sorted:
            # ToDo: refactor Table definition
            label = property.get_description_page_link() if clickable_links else property.label
            table_str += "!style=\"text-align:left\" |" + label + "\n"
            table_str += "| {{#if:" + Template.template_arg(property.name) + "|" + Template.template_arg(property.name) + "|}}\n"
            table_str += "|-\n"
        return table_str + "|}"


class ListOf(SMWPart):
    """
    Provides methods to generate a List of page for a topic
    """

    def __init__(self, wikiRender=None):
        super().__init__("List of", wikiRender)

    @staticmethod
    def get_page_name(topic: Topic):
        return f"List of {topic.pluralName}"


class Form(SMWPart):
    """
    Provides methods to render a complete Form or parts of a Form.
    For more details see: https://www.mediawiki.org/wiki/Extension:Page_Forms
    """

    regexps = {
        'Regexp:NaturalNumber': {
            'regexp': "/^[0-9]+$!^$/",
            'message': 'Must be a Number',
            'or char': '!'
        }
    }

    def __init__(self, wikiRender=None):
        if wikiRender is not None:
            wikiRender.template_env.globals['Form'] = self
        super().__init__("Form", wikiRender)
        #self.template = "event_form.jinja"

    @staticmethod
    def get_page_name(topic: Topic):
        return f"Form:{topic.name}"

    @staticmethod
    def page_form_function(tag, **kwargs):
        """
        ToDo
        Args:
            tag: Type of the form function. e.g.: field, form, info, ...
            **kwargs: parameters of the form function
        Returns:

        """
        return "{{{" + tag + SMW.render_parameters(presence_is_true=True, **kwargs) + "}}}"

    @staticmethod
    def standard_input_tag(input, oneliner=True, **kwargs):
        """
        Renders standard input tag
        For more detail see: https://www.mediawiki.org/wiki/Extension:Page_Forms/Defining_forms#'standard_input'_tag
        Args:
            input: If list the standard input tag is generated for each item in the list with the given parameters. Otherwise, the standart input tag is generatde for the given input
            oneliner: If true result will beone string line. Otherwise, multiple standard input tags will result in multiple lines.
            **kwargs: parameters of the standard input tag. If the parameter contains whitespace escape it with underscore.
        Returns:
        """
        if isinstance(input, list):
            res = ""
            for tag in input:
                res += Form.standard_input_tag(tag, oneliner, **kwargs)
            return res
        else:
            postfix = "" if oneliner else "\n"
            possible_input_tags = ["save", "preview", "save and continue", "changes", "summary", "minor edit", "watch", "cancel"]
            if input not in possible_input_tags:
                tags = ','.join(possible_input_tags)
                raise AttributeError(f"Form standard input \"{input}\" was given but must be one of the following values:{tags}")
            return Form.page_form_function(tag="standard input", **{input:True, **kwargs}) + postfix

    @staticmethod
    def forminput(**kwargs):
        """
        Renders a forminput parser function with the given input
        For more details see: https://www.mediawiki.org/wiki/Extension:Page_Forms/Linking_to_forms#Using_#forminput
        Args:
            **kwargs: parameters of the forminput
        Returns:
        """
        return SMW.parser_function(function_name="forminput", presence_is_true=True, **kwargs)

    @staticmethod
    def formlink(**kwargs):
        """
        Renders a formlink parser function with the given input
        For more details see: https://www.mediawiki.org/wiki/Extension:Page_Forms/Linking_to_forms#Using_#formlink
        Args:
            **kwargs: parameters of the formlink
        Returns:

        """
        return SMW.parser_function(function_name="formlink", presence_is_true=True, **kwargs)

    @staticmethod
    def field(property: Property):
        """
        Get the form field of the given property
        Args:
            property: property for which the field should be rendered
        Returns:
            Returns a form field of given property
        """
        prop_map = {"inputType": "input_type",
                    "placeholder": "placeholder",
                    "defaultValue": "default",
                    "values_from": "values_from",
                    "uploadable": "uploadable",
                    "primaryKey": "unique",
                    "allowedValues": "values"}
        parameters = {}
        for prop in prop_map.keys():
            if prop == "inputType":
                if property.__dict__[prop] is None:
                    property.__dict__[prop] = "text"
                elif property.__dict__[prop].lower() == "textarea":
                    parameters["editor"] = "wikieditor"
            if prop in property.__dict__ and property.__dict__[prop] is not None:
                if prop == "values_from":
                    # values_from contains data like "concept=Country" so we add an auxiliary property "values from concept" with the value "Country"
                    vals = property.__dict__[prop].split("=")
                    parameters[f"{prop_map[prop]} {vals[0]}" ] = vals[1]
                else:
                    parameters[prop_map[prop]] = property.__dict__[prop]
                if prop == "inputType" and property.__dict__[prop] == "regexp":
                    parameters.update(Form.regexps.get(property.regexp))
        return Form.page_form_function(tag="field", **{property.name: True, **parameters})

    @staticmethod
    def form_table(label: str, properties: list, is_collapsible=True, ):
        """
        Generate a form table based on the given parameters.
        Note: the result is not a complete form (start and end tags are missing)   # ToDo: Since the result is not complete form consider renaming this method to field_table
        Args:
            label: Headline of the table
            properties: List of properties for which a field should be added to the form table
            is_collapsible: If true the table is collapsible. Other wise the table will be static
        Returns:
            Returns a table containing form fields based on the given properties
        """
        # Define styling
        table_style = {"css_class": "wikitable", "style": "width:100%"}
        if is_collapsible:
            table_style["css_class"] += " mw-collapsible"
        field_label_style = {"style": "text-align: left"}
        # Define the table
        table = Table(**table_style)
        # Add table headline
        if label:
            table.add_row().add_cell(content=label, colspan=2)
        # Add form field for each given property
        sortBySortPos = lambda property: 99999 if "sortPos" not in property.__dict__ or property.__dict__.get("sortPos") is None else int(property.sortPos)
        properties_sorted = sorted(properties, key=sortBySortPos)
        for property in properties_sorted:
            property_row = table.add_row()
            # Add field label
            field_label = property.get_description_page_link()
            if property.mandatory:
                field_label += "<span style=\"color:#f00\">*</span>"
            field_label += ":"
            property_row.add_cell(content=field_label, is_header=True, **field_label_style)
            # Add field
            property_row.add_cell(content=Form.field(property))
        return table.render()



class Query:
    """
    SMW Query
    """

    def __init__(self,
                 mode="ask",
                 mainlabel=None,
                 limit: int = None,
                 offset: int = None,
                 searchlabel: str = None,
                 outro: str = None,
                 intro: str = None,
                 default: str = None,
                 format: str = None,
                 headers: str = None,
                 order: str = None,
                 template: str = None):
        """
        Intit a Query with the given parameters
        Args:
            mode: ask for multi page query / show for single page query
            mainlabel: title of the first column
            limit: maximal number of pages selected
            offset: where to start selecting pages
            searchlabel: text for continuing the search
            outro: text that is appended to the output, if at least some results exist
            intro: initial text that prepends the output, if at least some results exist
            default: if, for any reason, the query returns no results, this will be printed instead
            format: output format  of the query
            headers: Show headers (with links), plain headers (just text) or hide them. show is default. Possible values show, hide and, plain.
            order: defines how results should be ordered, ascending is the default. Possible values ascending/asc, descending/desc/reverse, random/rand.
        """
        self.mode = mode
        self.page_selections = []
        self.printout_statements = []
        self.mainlabel = mainlabel
        self.limit = limit
        self.offset = offset
        self.searchlabel = searchlabel
        self.outro = outro
        self.intro = intro
        self.default = default
        self.format = format
        self.headers = headers
        self.order = order
        self.sort = None
        self.template= template
        self.parameters = ["limit", "offset", "format", "searchlabel", "outro", "intro", "default", "headers", "sort", "order", "template"]

    def render(self, oneliner=True):
        """
        Render this query to the string format of an SMW query
        Args:
            oneliner: If True the query is rendered into one line. Otherwise the query is rendered in a prettier format.
        Returns:
            Returns SMW query of this query object
        """
        separator = ""
        if not oneliner:
            separator = "\n"
        query = "{{#" + self.mode + ":"
        for selector in self.page_selections:
            query += f"[[{selector}]]"
        query += separator
        if self.mainlabel is not None:
            query += f"|mainlabel={self.mainlabel}{separator}"
        for printout_statement in self.printout_statements:
            query += f"|{printout_statement.render()}{separator}"
        for parameter in self.parameters:
            if self.__dict__[parameter] is not None:
                query += f"|{parameter}={self.__dict__[parameter]}{separator}"
        return query + "}}"

    def printout(self, property, label=None):
        """
        Add a printout statement to the query. Results in querying the given property name.
        Args:
            property: Name of the property that should be queried OR Property object that should be queried
            label: Label that should be displayed at the result

        Returns:
            Returns the query itself
        """
        if isinstance(property, str):
            self.printout_statements.append(self.PrintoutStatement(property, label))
        else:
            self.printout_statements.append(self.PrintoutStatement(property.get_pageTitle(withNamespace=False), label))
        return self

    def select(self, selector):
        """
        Add page selector to the query
        Args:
            selector:

        Returns:
            Returns the query itself
        """
        self.page_selections.append(selector)
        return self

    def printout_list_of_properties(self, properties: list, withNamespace=False):
        """
        Add a printout statement for each property of the given properties
        Args:
            properties: All properties that should be queried
        Returns:
            The query object itself
        """
        for property in properties:
            self.printout(property.get_pageTitle(withNamespace), property.label)
        return self

    def sort_by(self, *properties):
        """
        Set the sort parameter of the query. If already set it overwrites the existing parameter.
        Args:
            *properties:

        Returns:

        """
        self.sort = ", ".join(properties)
        return self

    class PrintoutStatement:
        """Printout statement of a SMW query"""

        def __init__(self, property_name, label=None):
            """
            Init the printout statement
            Args:
                property_name: Name of the property that should be queried
                label: Label that should be displayed at the result
            """
            self.property_name = property_name
            self.label = label

        def render(self):
            """
            Render this PrintoutStatement to the string format of an printout statement
            Returns:

            """
            label = f"={self.label}" if self.label else ""
            return f"?{self.property_name}{label}"


class Table:
    """
    Experimental Class to test if generating MediaWiki tables can be streamlined
    Provides methods to generate a MediaWiki table
    For more information see https://www.mediawiki.org/wiki/Help:Tables
    """

    def __init__(self, css_class: str=None, style: str=None ):
        """
        Initialize the table with the given style/ style class
        Args:
            css_class: css class the table should have
            style: individual css styles that should be applied to the class such as "width:100%"
        """
        self.rows = []
        self.css_class = css_class
        self.style = style

    def render(self):
        """Renders this table as wiki table"""
        res =  "{| " + Table.render_parameter(self.css_class, self.style) +"\n"
        for row in self.rows:
            res += row.render()
        return res + "|}"

    def add_row(self, css_class: str=None, style: str=None, **kwargs):
        """
        Add a row to this table with the given styling configs
        Args:
            css_class: css class that the row should have
            style: individual css style that should be applied dto the row
            **kwargs: Additional parameter that should be set
        Returns:
            Returns the row object that was added to the table.
        """
        row = self.Row(css_class, style, **kwargs)
        self.rows.append(row)
        return row

    @staticmethod
    def render_parameter(css_class: str=None, style: str=None, **kwargs):
        """Render given parameter to the syntax of wiki tables"""
        res = ""
        res += f"class=\"{css_class}\" " if css_class else ""
        res += f"style=\"{style}\" " if style else ""
        for key, value in kwargs.items():
            if value:
                res += f"{key}=\"{value}\" "
        return res

    class Row:
        """
        Representing a row of a wiki table
        """

        def __init__(self, css_class: str=None, style: str=None):
            """
            Initialize table row with the given row style
            Args:
                css_class: css class of this row
                style: individual css style that should be applied to the row
            """
            self.cells = []
            self.css_class = css_class
            self.style = style

        def render(self):
            """Render this row as wiki table row"""
            res = "|-" + Table.render_parameter(self.css_class, self.style) +"\n"
            for cell in self.cells:
                res += cell.render() + "\n"
            return res[:-1] + "\n"

        def add_cell(self, content, is_header=False, colspan=None, css_class: str=None, style: str=None):
            """
            Add a cell to this row. The order of inserting cells corresponds to the ordering in the rendered table.
            First cell added is first column in the table.
            Args:
                content: Content of the cell
                is_header: True if this cell is a header cell.
                colspan: span of the cell
                css_class: css class of this cell
                style: individual css style that should be applied to the cell
            Returns:

            """
            cell = self.Cell(content, is_header, colspan, css_class, style)
            self.cells.append(cell)

        class Cell:
            """
            Representing a cell of a wiki table
            """
            def __init__(self, content, is_header=False, colspan: int=None, css_class: str=None, style: str=None):
                """
                Initialize table cell with the given content and style configs
                Args:
                    content: Content of the cell
                    is_header: True if this cell is a header cell.
                    colspan: span of the cell
                    css_class: css class of this cell
                    style: individual css style that should be applied to the cell
                """
                self.content = content
                self.is_header = is_header
                self.colspan = colspan
                self.css_class = css_class
                self.style = style

            def render(self):
                """Render this cell as wiki table cell"""
                separator = "!" if self.is_header else "|"
                config = Table.render_parameter(self.css_class, self.style, colspan=self.colspan)
                config += " |" if config else ""
                return f"{separator}{config}{self.content}"




# ! important
# see https://stackoverflow.com/a/27868004/1497139
from setuptools import setup, find_packages

# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='wikirender',
    version='0.0.13',
    packages=['wikifile','templates','templates/macros', 'wikifile/resources/metamodel'],
    classifiers=[
            'Programming Language :: Python',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9'
    ],

    install_requires=[
      'wikitextparser',
      'jinja2',
      'pylodstorage'
    ],
    entry_points={
      'console_scripts': [
        'wikirender = wikifile.wikiRender:main_module_call',
        'wikiextract = wikifile.wikiExtract:main_module_call',
      ],
    },
    package_data={'templates': ['*.jinja', 'templates/*.jinja','templates/macros/*.jinja'],'templates/macros': ['*.jinja', 'templates/*.jinja','templates/macros/*.jinja'], "wikifile/resources/metamodel":['*.json']},
    author='Tim Holzheim',
    maintainer='Tim Holzheim',
    url='https://github.com/tholzheim/wikirender',
    license='Apache License',
    description='Convert Json data into wiki files or update existing files. Provides different templates to generate wiki files',
    long_description=long_description,
    long_description_content_type='text/markdown'
)



import os
import wikitextparser as wtp
from mwclient.page import Page
from wikitextparser import Template


class WikiFile:
    '''
    Provides methods to modify, query, update and save files that contain wiki text markup.
    see https://en.wikipedia.org/wiki/Help:Wikitext
    '''

    def __init__(self, name, path, wiki_render, wikiText: str = None, debug=False):
        """

        Args:
            name: Name of the file
            path: Path to file location
            wiki_render:
            wikiText: WikiPage content as string. If defined the file content will loaded and this value will be used instead
        """
        self.file_name = name
        self.file_path = path
        self.debug = debug
        if wikiText is None:
            self.wikiText = self.get_wikiText(self.file_name, self.file_path)
        else:
            self.wikiText = wtp.parse(wikiText)
        self.wiki_render = wiki_render

    def save_to_file(self, overwrite=False):
        """
        Save the given data in a file.
        
        Args:
            overwrite(bool): If True existing files will be over written and if the path does not exist the missing folder will be created. Otherwise, only missing files will be saved.
  
        """
        wiki_file_path = WikiFile.get_wiki_path(self.file_path, self.file_name)
        mode = "w"
        if overwrite:
            mode = 'w'
            os.makedirs(os.path.dirname(wiki_file_path), exist_ok=True)
        else:
            if os.path.isfile(wiki_file_path):
                # file already exists
                if self.debug:
                    print(f"File {wiki_file_path} exists \t-> Generated page not saved. To save also existing pages use -f to overwrite them.", )
                    return
        with open(wiki_file_path, mode=mode) as f:
            f.write(str(self.wikiText))
        if self.debug:
            print(f"{self.file_name} saved to {wiki_file_path}")

    def get_wikiText(self, name, path):
        """
        find the wiki file by name and return it as parsed object.
        Args:
            name: Name of the file
            path: Path to the file

        Returns:
            WikiText object
        """
        fname = WikiFile.get_wiki_path(path, name)
        if os.path.isfile(r"{}".format(fname)):
            with open(fname, mode='r') as file:
                page = file.read()
                parsed_page = wtp.parse(page)
                return parsed_page
        return None

    def get_template(self, template_name: str):
        """
        Returns the template
        
        Args:
            template_name(str): the name of the template to extract

        Returns:

        """
        if self.wikiText is None or self.wikiText.templates is None:
            # Wikifile has no templates
            return None
        for template in self.wikiText.templates:
            name = WikiFile.get_template_name(template.name)
            if name == WikiFile.get_template_name(template_name):
                return template
        return None

    @staticmethod
    def get_template_name(template_name: str):
        """
        Removes the whitespace around the template name and the line break if present.
        Example: " Event \n" -> "Event"

        Args:
            template_name: raw template name that should be normalized

        Returns:
            normalized template name.
        """
        name = template_name.replace('\n', '')
        name = name.lstrip()
        name = name.rstrip()
        return name

    def update_template(self, template_name: str, args:dict, overwrite=False, prettify:bool=False):
        """
        Updates the given template the values from the given dict args.
        If force is set to True existing values will be overwritten.
        ToDo: Currently unhandled is the behavior if multiple templates with the same name exist
        Args:
            template_name(str): name of the template that should be updated
            args(dict): Dict containing the arguments that should be set. key=argument name, value=argument value
            overwrite(bool): If True existing values will be overwritten
            prettify(bool): If True new values will be added into a new line

        Returns:
            Nothing
        """
        template = self.get_template(template_name)
        self.update_arguments(template, args, overwrite, prettify)

    @staticmethod
    def update_arguments(template: Template, args:dict, overwrite=False, prettify:bool=False):
        """
        Updates the arguments of the given template with the values of the given dict (args)

        Args:
            template: Template that should be updated
            args(dict): Dict containing the arguments that should be set. key=argument name, value=argument value
            overwrite(bool): If True existing values will be overwritten
            prettify(bool): If True new values will be added into a new line

        Returns:
            Nothing
        """
        for key, value in args.items():
            if template.has_arg(key):
                # update argument
                if overwrite:
                    template.del_arg(key)
                    template.set_arg(key, value)
                else:
                    pass
            else:
                if value is not None:
                    postfix=""
                    if prettify:
                        postfix="\n"
                    template.set_arg(key, value+postfix, preserve_spacing=False)

    def add_template(self, template_name: str, data: dict, overwrite=False, prettify:bool=False):
        """
        Adds the given data as template with the given name to this wikifile.
        If the template is already present only the arguments are updated (the values are only overwritten if overwrite
        is set to True)
        ToDo: Currently unhandled is the behavior if multiple templates with the same name exist
        Args:
            template_name(str): Name of the template the data should be inserted in
            data(dict): Data that should be saved in form of a template
            overwrite(bool): If true existing values are overwritten. Otherwise onl missing values are added. Default: False
            prettify: If true each newly added value will be placed in a new line. Oterwise values are added in the same line. Default is false.

        Returns:
            Nothing
        """
        template = self.get_template(template_name)
        if template is None:
            # create template
            self.wikiText.insert(0, self.wiki_render.render_template(template_name, data))
        else:
            WikiFile.update_arguments(template, data, overwrite, prettify)

    def extract_template(self, name: str):
        """
        Extracts the template data and returns it as dict
        ToDo: Currently unhandled is the behavior if multiple templates with the same name exist
        
        Args:
            name: name of the template that should be extracted

        Returns:
            Returns template content as dict
        """
        template = self.get_template(name)
        if template is None:
            return None
        res = {}
        for arg in template.arguments:
            value = arg.value.strip()
            if value.endswith("\n"):
                value=value[:-1]
            res[arg.name.strip()] = value
        return res

    def setPage(self, page:Page):
        self.pageRef=page

    def getPage(self)->Page:
        if 'pageRef' in self.__dict__:
            return self.__dict__["pageRef"]

    def getPageTitle(self):
        return self.file_name

    @staticmethod
    def get_wiki_path(path: str, name: str):
        return f"{path}/{name}.wiki"

    def __str__(self) -> str:
        return str(self.wikiText)

